use crate::pattern::{AbsCoord, Pattern};
use colored::{Color, Colorize};
use std::fmt;

/// A Field is a list of available (empty) coordinates.
#[derive(Debug, Clone)]
pub struct Field {
    width: u32,
    height: u32,
    pub empty_coords: Vec<AbsCoord>,
    patterns: Vec<(AbsCoord, Pattern)>,
}

impl Field {
    pub fn new(width: u32, height: u32) -> Self {
        let mut empty_coords = Vec::new();
        for y in 0..height {
            for x in 0..width {
                empty_coords.push(AbsCoord { x, y });
            }
        }
        Self {
            width,
            height,
            empty_coords,
            patterns: Vec::new(),
        }
    }

    pub fn add_pattern(&mut self, location: &AbsCoord, pattern: &Pattern) -> bool {
        // Convert the pattern's relative coordinates to absolute coordinates
        let abs_coords: Option<Vec<_>> = pattern
            .coords
            .iter()
            .map(|rel_coord| location.add(rel_coord))
            .collect();
        match abs_coords {
            Some(abs_coords) => {
                if abs_coords
                    .iter()
                    .all(|abs_coord| self.empty_coords.contains(abs_coord))
                {
                    self.empty_coords
                        .retain(|empty_coord| !abs_coords.contains(empty_coord));
                    self.patterns.push((location.clone(), pattern.clone()));
                    true
                } else {
                    false
                }
            }
            None => false,
        }
    }

    pub fn fits(&self, location: &AbsCoord, pattern: &Pattern) -> bool {
        let abs_coords: Option<Vec<_>> = pattern
            .coords
            .iter()
            .map(|rel_coord| location.add(rel_coord))
            .collect();
        match abs_coords {
            Some(abs_coords) => {
                if abs_coords
                    .iter()
                    .all(|abs_coord| self.empty_coords.contains(abs_coord))
                {
                    true
                } else {
                    false
                }
            }
            None => false,
        }
    }
}

impl fmt::Display for Field {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let mut grid = vec![
            vec![' '.to_string().color(Color::White); self.width as usize];
            self.height as usize
        ];

        let colors = vec![
            Color::Red,
            Color::Green,
            Color::Yellow,
            Color::Blue,
            Color::Magenta,
            Color::Cyan,
            Color::White,
            Color::Black,
        ];
        for (index, (location, pattern)) in self.patterns.iter().enumerate() {
            let color = colors[index % colors.len()]; // This makes sure we loop back to the start if there are more patterns than colors.
            for rel_coord in &pattern.coords {
                let abs_coord = location.add(rel_coord).unwrap();
                grid[abs_coord.y as usize][abs_coord.x as usize] = '#'.to_string().color(color);
            }
        }

        for row in grid.iter() {
            for cell in row.iter() {
                write!(f, "{}", cell)?;
            }
            writeln!(f)?;
        }

        Ok(())
    }
}
